;       
        .xlist
;
;       Written By:     Kenn Gough
;                       Data Planning, Inc.
;                       Myrtle Beach, SC   USA
;
;                       kenngough@dataplaning.com
;
;                       (843) 215-0855 home
;                       (843) 241-6766 cell
;
;       128-Bit Math routines for add, subtract, multiply,
;       and divide.
;
;       This code was written for a request on the Masm32 Forum
;       and may be shared, modified or used as needed
;
;       This code was built using the QEditor, Project Menu,
;       Console Assemble & Link option
;
        .586
        .model  flat, stdcall
        option  casemap :none
        option  offset  :flat
;
        include \masm32\include\windows.inc
        include \masm32\include\kernel32.inc
        include \masm32\include\user32.inc
        include \masm32\include\winmm.inc
;
        includelib \masm32\lib\kernel32.lib
        includelib \masm32\lib\user32.lib
        includelib \masm32\lib\winmm.lib
;
;       These are the working 128 bit numbers
;       The bits are referenced from highest to lowest dd's
;       the lowest bits in the fourth double word
;
;               example 0,0,0,1 = 1
;                       0,0,1,0 = 65536
;
;       Due to bit rotation, these definations must remain in
;       the following order.  Do not insert in other variables
;       between them
;
.data
;
DGT_WRK dd      0,0,0,0                 ; work digit (128 bits)
DGT_TWO dd      0,0,0,0                 ; work digit two
DGT_ONE dd      0,0,0,0                 ; work digit one
;
RSLT_SN db      0
;
.code
;
Start:
;
;       The following were calls to the 128-bit math subroutine
;       and the results were reviewed with the debugger.
;
;       --------------------
;       Test Add Two Numbers
;       --------------------
;
        call    Sub_Zero_128                    ; clear all numbers
;
        mov     eax, 100                        ; first number
;
        mov     DGT_ONE+12, eax
;
        mov     eax, 100                        ; second number
;
        mov     DGT_TWO+12, eax
;
        call    Sub_Add_Two_128                 ; add together
;
        mov     eax, DGT_ONE                    ; get               (Sum)
        mov     ebx, DGT_ONE+4                  ;   answer
        mov     ecx, DGT_ONE+8                  ;     to
        mov     edx, DGT_ONE+12                 ;       view
;
;       -------------------------
;       Test Subtract Two Numbers
;       -------------------------
;
        call    Sub_Zero_128                    ; clear all numbers
;
        mov     eax, 100                        ; first number
;
        mov     DGT_ONE+12, eax
;
        mov     eax, 36                         ; second number
;
        mov     DGT_TWO+12, eax
;
        call    Sub_Subtract_Two_128            ; subtraact 2nd from 1st
;
        mov     eax, DGT_ONE                    ; get          (Difference)
        mov     ebx, DGT_ONE+4                  ;   answer
        mov     ecx, DGT_ONE+8                  ;     to
        mov     edx, DGT_ONE+12                 ;       view
;
;       -------------------------
;       Test Multiply Two Numbers
;       -------------------------
;
        call    Sub_Zero_128                    ; clear all numbers
;
        mov     eax, 10                         ; first number
;
        mov     DGT_ONE+12, eax
;
        mov     eax, 20                         ; second number
;
        mov     DGT_TWO+12, eax
;
        call    Sub_Multiply_Two_128            ; multiply together
;
        mov     eax, DGT_ONE                    ; get           (Product)
        mov     ebx, DGT_ONE+4                  ;   answer
        mov     ecx, DGT_ONE+8                  ;     to
        mov     edx, DGT_ONE+12                 ;       view
;
;       ------------------------
;       Test Divide Two Numbers
;       ------------------------
;
        call    Sub_Zero_128                    ; clear all numbers
;
        mov     eax, 102                        ; first number
;
        mov     DGT_ONE+12, eax
;
        mov     eax, 20                         ; second number
;
        mov     DGT_TWO+12, eax
;
        call    Sub_Divide_Two_128              ; Divide 2nd Number into 1st
;
        mov     eax, DGT_ONE                    ; get             (Quotient)
        mov     ebx, DGT_ONE+4                  ;   answer
        mov     ecx, DGT_ONE+8                  ;     to
        mov     edx, DGT_ONE+12                 ;       view
;
        mov     eax, DGT_TWO                    ; get
        mov     ebx, DGT_TWO+4                  ;   answer       (Remainder)
        mov     ecx, DGT_TWO+8                  ;     to
        mov     edx, DGT_TWO+12                 ;       view
;
        invoke  ExitProcess, \                  ; exit program
                        NULL
;
;------------------------------------------------
;       Add Two 128-Bit Numbers
;------------------------------------------------
;
;       Routine adds 2nd number to first
;
;       On Entry:
;               DGT_ONE = first number
;               DGT_TWO = second number
;
;       On Exit:
;               DGT_ONE = sum of two numbers
;       
Sub_Add_Two_128:
;
        clc                                     ; clear carry flag
;
        mov     eax, DGT_TWO+12                 ; lowest dd
        adc     DGT_ONE+12, eax                 ; add 2nd number to 1st
;
        mov     eax, DGT_TWO+8                  ; second dd
        adc     DGT_ONE+8, eax                  ; add 2nd number to 1st
;
        mov     eax, DGT_TWO+4                  ; third dd
        adc     DGT_ONE+4, eax                  ; add 2nd number to 1st
;
        mov     eax, DGT_TWO                    ; fourth dd
        adc     DGT_ONE, eax                    ; add 2nd number to 1st
;
        ret                                     ; exit routine
;
;------------------------------------------------
;       Subtract Two 128-Bit Numbers
;------------------------------------------------
;
;       Routine subtracts 2nd number from the first
;
;       On Entry:
;               DGT_ONE = first number
;               DGT_TWO = second number
;
;       On Exit:
;               DGT_ONE = difference of the two numbers
;       
Sub_Subtract_Two_128:
;
        mov     esi, offset DGT_TWO             ; 2nd number location
;
        call    Sub_Negate_128                  ; compliments number
;
        call    Sub_Add_Two_128                 ; not add to subtract
;
        ret                                     ; exit routine
;
;------------------------------------------------
;       Negate A Number
;------------------------------------------------
;
;       Routine compliments the indicated number
;       to change the its sign
;
;       On Entry:
;               esi = points to number to reverse sign
;
;       On Exit:
;               sign  reverse on indicated nnumber
;       
Sub_Negate_128:
;
        add     esi, 12                         ; point to last dd
;
        clc                                     ; clear carry flag
;
        mov     eax, 0                          ; to subtract from zero
        sbb     eax, [esi]                      ; subtract with borrow
        mov     [esi], eax                      ; update word
;
        dec     esi                             ; reduce
        dec     esi                             ;   without
        dec     esi                             ;     changing
        dec     esi                             ;       carry flag
;
        mov     eax, 0                          ; to subtract from zero
        sbb     eax, [esi]                      ; subtract with borrow
        mov     [esi], eax                      ; update word
;
        dec     esi                             ; reduce
        dec     esi                             ;   without
        dec     esi                             ;     changing
        dec     esi                             ;       carry flag
;
        mov     eax, 0                          ; to subtract from zero
        sbb     eax, [esi]                      ; subtract with borrow
        mov     [esi], eax                      ; update word
;
        dec     esi                             ; reduce
        dec     esi                             ;   without
        dec     esi                             ;     changing
        dec     esi                             ;       carry flag
;
        mov     eax, 0                          ; to subtract from zero
        sbb     eax, [esi]                      ; subtract with borrow
        mov     [esi], eax                      ; update word
;
        ret                                     ; exit routine
;
;------------------------------------------------
;       Set Resulting Sign
;------------------------------------------------
;
;       Routine test the sign of the two numbers,
;       for multiplying or dividing, reverses any
;       negatives, and sets what the resulting
;       sign should be
;
;       On Entry:
;               DGT_ONE = first number
;               DGT_TWO = second number
;
;       On Exit:
;               DGT_ONE & DGT_TWO both positive numbers
;               RSLT_SN indicates resulting sign 
;       
Sub_Set_Sign_Result:
;
        xor     al, al                          ; clear for work
;
        mov     RSLT_SN, al                     ; clear result sign
;
;
;       -----------------
;       Test First Number
;       -----------------
;
        mov     eax, DGT_ONE                    ; high order dd
;
        test    eax, 080000000h                 ; sign bit
        je      @f                              ;   not negative
;
        mov     esi, offset DGT_ONE             ; 1st number address
;
        call    Sub_Negate_128                  ; reverse sign
;
        inc     RSLT_SN                         ; count negative's
;
;       ------------------
;       Test Second Number
;       ------------------
@@:
        mov     eax, DGT_TWO                    ; high order dd
;
        test    eax, 080000000h                 ; sign bit
        je      @f                              ;   not negative
;
        mov     esi, offset DGT_TWO             ; 2nd number address
;
        call    Sub_Negate_128                  ; reverse sign
;
        inc     RSLT_SN                         ; count negative's
@@:
        ret                                     ; exit routine
;
;------------------------------------------------
;       Multiply Two Signed 128-Bit Numbers
;------------------------------------------------
;
;       Routine multiplies two 128-bit numbers and
;       returns a 128-bit signed number.
;
;       This is an add and rotate routine.
;
;       On Entry:
;               DGT_ONE = first number
;               DGT_TWO = second number
;
;       On Exit:
;               DGT_ONE = product of the two numbers
;               correct sign is set
;       
Sub_Multiply_Two_128:
;
;       ---------------------------------------------
;       Calculate Resulting Sign / Make Both Positive
;       ---------------------------------------------
;
        call    Sub_Set_Sign_Result             ; set sign / make positive
;
;       --------------------------------
;       Move Second Number To Work Digit
;       --------------------------------
;
        mov     esi, offset DGT_TWO             ; second 128-bit number
        mov     edi, offset DGT_WRK             ; work 128-bit number
        mov     ecx, 4                          ; dd's to move
rep     movsd                                   ; copy number
;
;       ------------------------------
;       Clear 2nd Number Area For Work
;       ------------------------------
;
        mov     edi, offset DGT_TWO             ; really already there
        xor     eax, eax                        ; clear for work
        mov     ecx, 4                          ; dd's to clear
rep     stosd                                   ; fill with zeroes
;
        mov     ecx, 128                        ; bits in numbers
;
;       ------------------------------------------------------
;       How This Works:
;
;       The original first number is still in DGT_ONE.
;       The second number is in the work number.  DGT_TWO is
;       cleared to build the answer.  The low bit of
;       the first number is tested to see if to be multiplied
;       by the second (0 or 1).  if yes, the second number is
;       added to the working answer.  Finally both the working
;       answer and the original first number are shifted right
;       a bit.  Testing continues with next bit.  The final
;       product answer ends up in the original first number
;       location as the first number is shifted off.
;       ------------------------------------------------------
;
Sub_Multiply_128_B:
;
        push    ecx                             ; save bit count
;
        mov     eax, DGT_ONE+12                 ; lowest dd of 1st number
;
;       --------------------
;       Test To Multiply Bit
;       --------------------
;
        test    al, 1                           ; do we multiply?
        je      @F                              ;   no
;
;       --------------------
;       Add To Multiply By 1
;       --------------------
;
        clc                                     ; clear carry flag
;
        mov     eax, DGT_WRK+12                 ; multiplying by
        adc     DGT_TWO+12, eax                 ;   add to multiply by 1
;
        mov     eax, DGT_WRK+8                  ; multiplying by
        adc     DGT_TWO+8, eax                  ;   add to multiply by 1
;
        mov     eax, DGT_WRK+4                  ; multiplying by
        adc     DGT_TWO+4, eax                  ;   add to multiply by 1
;
        mov     eax, DGT_WRK                    ; multiplying by
        adc     DGT_TWO, eax                    ;   add to multiply by 1
;
@@:
;
;       -----------------------------------
;       Now Rotate The Working Product Thru
;       The First Number (All Shifts Right)
;       -----------------------------------
;
        mov     esi, offset DGT_TWO             ; working product address
        mov     edi, esi                        ; make both the same
;
        mov     ecx, 8                          ; total dd's (both numbers)
;
        clc                                     ; clear carry for rotate thru
;
@@:
        lodsd                                   ; get a double word
        rcr     eax, 1                          ; rotate right thru carry
        stosd                                   ; update double word
;
        loop    @B                              ; loop thru both dd's
;
;       --------------------------
;       Restore 128 bit loop count
;       --------------------------
;
        pop     ecx                             ; restore main loop count
;
        loop    Sub_Multiply_128_B              ; main loop
;
;       --------------------------------
;       Multiply Done - Set Correct Sign
;       --------------------------------
;
        mov     al, RSLT_SN                     ; get resulting sign
;
        test    al, 1                           ; to be positive?
        je      @F                              ;   no
;
        mov     esi, offset DGT_ONE             ; number to reverse sign
;
        call    Sub_Negate_128                  ; call reverse sign routine
;
@@:
        ret                                     ; exit routine
;
;------------------------------------------------
;       Divide Two Signed 128-Bit Numbers
;------------------------------------------------
;
;       Routine divides the second 128-bit numbers
;       into the first.  The remainder is left in the
;       second nuumber.
;
;       This is a rotate and subtract routine.
;
;       If either number is zero, both the quotient and
;       remainder are set to zero
;
;       On Entry:
;               DGT_ONE = first number
;               DGT_TWO = second number
;
;       On Exit:
;               DGT_ONE = quotient of the divide.
;
;               the remainder is DGT_TWO
;
;               correct sign is set on quotient
;               remainder is positive
;       
Sub_Divide_Two_128:
;
;       ---------------------------------------------
;       Calculate Resulting Sign / Make Both Positive
;       ---------------------------------------------
;
        call    Sub_Set_Sign_Result             ; set sign / make positive
;
;       ------------------------------------------------------
;       How This Works:
;
;       The first number is shifted left into the 2nd.  If the
;       2nd number can be subtracted from current value, a low
;       order bit is set, constructing the final answer.  Rather
;       than comparing to see if can subtract, we subtract and
;       test the resulting sign.  If it is negative, we should
;       not subtract, and the 2nd digit is added back.  Scaling
;       is done by testing the high order dd's to reduce the
;       number of subtract/add/rotates that are executed
;       ------------------------------------------------------
;
;       --------------------------------
;       Test First Number For All Zeroes
;       --------------------------------
;
        xor     edx, edx                        ; clear for work
;
        mov     esi, offset DGT_ONE             ; address first number
;
        mov     ecx, 4                          ; dd's to test
;
@@:
        lodsd                                   ; get a dd
        or      edx, eax                        ; put together
        loop    @B                              ; to get all
;
        or      edx, edx                        ; all zero ?
        je      Sub_Divide_128_B                ;   yes
;
;       ---------------------------------
;       Test Second Number For All Zeroes
;       ---------------------------------
;
        xor     edx, edx                        ; clear for work
;
        mov     esi, offset DGT_TWO             ; address first number
;
        mov     ecx, 4                          ; dd's to test
;
@@:
        lodsd                                   ; get a dd
        or      edx, eax                        ; put together
        loop    @B                              ; to get all
;
        or      edx, edx                        ; all zero ?
        jne     @F                              ;   no - continue
;
;       --------------------------
;       Clear Both Numbers To Zero
;       --------------------------
;
Sub_Divide_128_B:
;
        xor     eax, eax                        ; clear for work
        mov     edi, offset DGT_TWO             ; address of "numbers"
        mov     ecx, 8                          ; dd's to clear
;
rep     stosd                                   ; clear both numbers
;
        ret                                     ; early exit
;
;       --------------------------------
;       Move Second Number To Work Digit
;       --------------------------------
@@:
        mov     esi, offset DGT_TWO             ; second 128-bit number
        mov     edi, offset DGT_WRK             ; work 128-bit number
        mov     ecx, 4                          ; dd's to move
rep     movsd                                   ; copy number
;
;       ------------------------------
;       Clear 2nd Number Area For Work
;       ------------------------------
;
        mov     edi, offset DGT_TWO             ; really already there
        xor     eax, eax                        ; clear for work
        mov     ecx, 4                          ; dd's to clear
rep     stosd                                   ; fill with zeroes
;
;       ----------------
;       Set Up To Divide
;       ----------------
;
        mov     ecx, 128                        ; bits in numbers
;
        call    Sub_Divide_Scaling              ; shift divide-into left
;
;       ------------------
;       Now The Fun Begins
;       ------------------
;
Sub_Divide_128_D:
;
;       ------------------------------
;       Rotate Both Numbers Left A Bit
;       ------------------------------
;
        push    ecx                             ; save bit
;
        clc                                     ; clear carry
;
        rcl     DGT_ONE+12, 1                   ; high
        rcl     DGT_ONE+8, 1                    ;   bits
        rcl     DGT_ONE+4, 1                    ;     passed
        rcl     DGT_ONE, 1                      ;       thru
        rcl     DGT_TWO+12, 1                   ;         carry
        rcl     DGT_TWO+8, 1                    ;           flag
        rcl     DGT_TWO+4, 1                    ;
        rcl     DGT_TWO, 1                      ;
;
;       -------------------------------
;       Attemp To Subtract (Bit Divide)
;       -------------------------------
;
        clc                                     ; clear carry flag
;
        mov     eax, DGT_WRK+12                 ; divide by word
        sbb     DGT_TWO+12, eax                 ;   subtract with borry
;
        mov     eax, DGT_WRK+8                  ; divide by word
        sbb     DGT_TWO+8, eax                  ;   subtract with borry
;
        mov     eax, DGT_WRK+4                  ; divide by word
        sbb     DGT_TWO+4, eax                  ;   subtract with borry
;
        mov     eax, DGT_WRK                    ; divide by word
        sbb     DGT_TWO, eax                    ;   subtract with borry
;
;       ---------------------------
;       Test For Sucessful Subtract
;       ---------------------------
;
        mov     eax, DGT_TWO                    ; dd with sign bit
;
        test    eax, 080000000h                 ; test sign bit
        je      Sub_Divide_128_E                ;   was able to "divide"
;
;       ------------------------
;       Divide Failed - Add Back
;       ------------------------
;
        clc                                     ; clear carry flag
;
        mov     eax, DGT_WRK+12                 ; divide by word
        adc     DGT_TWO+12, eax                 ;   add back with carry
;
        mov     eax, DGT_WRK+8                  ; divide by word
        adc     DGT_TWO+8, eax                  ;   add back with carry
;
        mov     eax, DGT_WRK+4                  ; divide by word
        adc     DGT_TWO+4, eax                  ;   add back with carry
;
        mov     eax, DGT_WRK                    ; divide by word
        adc     DGT_TWO, eax                    ;   add back with carry
;
        jmp     Sub_Divide_128_F                ; skip answer update
;
Sub_Divide_Jump_Back:
;
        jmp     Sub_Divide_128_D
;        
Sub_Divide_128_E:
;
        inc     DGT_ONE+12                      ; update answer
;
Sub_Divide_128_F:
;
;       -----------------------------
;       Finished Pass - Loop Til Done
;       -----------------------------
;
        pop     ecx                             ; restore bit loop count
;
        loop    Sub_Divide_Jump_Back            ; continue
;
;       -----------------------------
;       Divide Complete - Fix Up Sign
;       -----------------------------
;
        mov     al, RSLT_SN                     ; sign to be
;
        test    al, 1                           ; odd number of signs
        je      @F                              ;   no
;
;       ---------------------
;       Reverse Quotient Sign
;       ---------------------
;
        mov     esi, offset DGT_ONE             ; address of quotient
;
        call    Sub_Negate_128                  ; reverse sign
;
@@:
;       ---------------
;       Good Exit Point
;       ---------------
;
        ret                                     ; and exit
;
;------------------------------------------------
;       Scaling For Divide
;------------------------------------------------
;
;       Routine "left justifies" the number to
;       divide into to reduce the bit rotation
;       passes.  The divide into number has
;       previously been tested for non-zero
;
;       On Entry:
;               DGT_ONE contains the number to
;               divide into
;
;               ecx contains the bit pass count
;
;       On Exit:
;               DGT_ONE is shifted left to byte
;               level and ecx has been reduced
;               accordingly
;
Sub_Divide_Scaling:
;
;       --------------------------
;       Scaling For High Zero DD's
;       --------------------------
@@:
        mov     eax, DGT_ONE                    ; high dd of divide into
;
        or      eax, eax                        ; all zeroes
        jne     @F                              ;   no - dd level done
;
;       ----------------------------------------------------
;       Shift Number One dd Left - Number Two Already Zeroes
;       ----------------------------------------------------
;
        mov     eax, DGT_ONE+4                  ; moving
        mov     DGT_ONE, eax                    ;   dd's left
;
        mov     eax, DGT_ONE+8                  ; moving
        mov     DGT_ONE+4, eax                  ;   dd's left
;
        mov     eax, DGT_ONE+12                 ; moving
        mov     DGT_ONE+8, eax                  ;   dd's left
;
        xor     eax, eax                        ; clear for work
;
        mov     DGT_ONE+12, eax                 ; clear dd of answer
;
        sub     ecx, 32                         ; reduce bits to process
;
        jne     @B                              ; loop til done
;
;       ---------------------------------------
;       Scaling Last DD by For High Zero Byte's
;       Significant bits now highest dd
;       ---------------------------------------
@@:
        mov     al, byte ptr DGT_ONE+3          ; high byte 4th dd in number
;
        or      al, al                          ; test zeroes
        jne     @F                              ;   no - significant bits
;
;       ----------------------------------------------------
;       Shift Number Left - Don't Forget Intel Storage Order
;       ----------------------------------------------------
;
        push    ecx                             ; save bit count
;
        mov     eax, DGT_ONE                    ; highes dd in number  (dd-1)
        mov     cl, 8                           ; bits to shift left
        shl     eax, cl                         ; shift off byte
;
        xor     edx, edx                        ; clear for work
        mov     dl, byte ptr DGT_ONE+7          ; high byte next dd
        or      eax, edx                        ; put together
        mov     DGT_ONE, eax                    ; put back
;
        mov     eax, DGT_ONE+4                  ; highes dd in number  (dd-2)
        mov     cl, 8                           ; bits to shift left
        shl     eax, cl                         ; shift off byte
;
        xor     edx, edx                        ; clear for work
        mov     dl, byte ptr DGT_ONE+11         ; high byte next dd
        or      eax, edx                        ; put together
        mov     DGT_ONE+4, eax                  ; put back
;
        mov     eax, DGT_ONE+8                  ; highes dd in number  (dd-3)
        mov     cl, 8                           ; bits to shift left
        shl     eax, cl                         ; shift off byte
;
        xor     edx, edx                        ; clear for work
        mov     dl, byte ptr DGT_ONE+15         ; high byte next dd
        or      eax, edx                        ; put together
        mov     DGT_ONE+8, eax                  ; put back
;
        mov     eax, DGT_ONE+12                 ; highes dd in number  (dd-4)
        mov     cl, 8                           ; bits to shift left
        shl     eax, cl                         ; shift off byte
;
        mov     DGT_ONE+12, eax                 ; put back
;
        pop     ecx                             ; restore bit-loop count
;
        sub     ecx, 8                          ; reduce by 8 bits
;
        jmp     @B                              ; test again
;
@@:
        ret                                     ; exit scaling
;
;------------------------------------------------
;       Misc - Clear Words Routine
;------------------------------------------------
;
;       Routine clears words for testing.
;
;       On Entry:
;               nothing passed
;
;       On Exit:
;               DGT_ONE, DGT_TWO, DGT_WRK are all zeroes
;
Sub_Zero_128:
;
        xor     eax, eax                        ; clear for work
;
        mov     edi, offset DGT_WRK             ; first of three numbers
;
        mov     ecx, 12                         ; dd's to clear
;
rep     stosd                                   ; fill with zeroes
;
        ret                                     ; and exit
;
end     Start

